# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


import urwid


class ListBoxWithMouseEvents(urwid.ListBox):
    def mouse_event(self, size, event, button, col, row, focus):
        if event == "mouse press":
            if button == 4:
                for _ in range(2):
                    self.keypress(size, "up")
                return True
            if button == 5:
                for _ in range(2):
                    self.keypress(size, "down")
                return True
        return super().mouse_event(size, event, button, col, row, focus)



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


import urwid

from picsh.controllers.cluster_selection_controller import ClusterSelectionController
from picsh.controllers.cluster_shell_controller import ClusterShellController
from picsh.controllers.node_panel_controller import NodePanelController
from picsh.controllers.receive_buffer_controller import ReceiveBufferController
from picsh.controllers.root_controller import RootController
from picsh.controllers.single_shell_controller import SingleShellController
from picsh.models.root_model import RootModel
from picsh.state_change_notifier import StateChangeNotifier
from picsh.views.cluster_selection_view import ClusterSelectionView
from picsh.views.cluster_shell_view import ClusterShellView
from picsh.views.node_panel_view import NodePanelView
from picsh.views.receive_buffer_view import ReceiveBufferView
from picsh.views.single_shell_view import SingleShellView


class App:
    def __init__(self, cluster_specs):
        self._cluster_specs = cluster_specs

    def run(self):
        urwid.set_encoding("utf8")

        palette = [
            ("separater", "white", "dark cyan"),
            ("body", "dark cyan", "black", "standout"),
            ("node_list", "dark gray", "black", "standout"),
            ("header_style", "dark cyan", "black"),
            ("footer", "black", "yellow"),
            ("footer_title", "white", "dark red"),
            ("footer_style", "black", "white"),
            ("reveal_focus", "black", "dark cyan", "standout"),
            ("node_text", "light gray", "black"),
            ("column_headers", "dark cyan", "black", "standout"),
            ("output_separater", "black", "dark gray", "standout"),
            ("cmdshell", "white", "black", "standout"),
        ]

        root_model = RootModel(cluster_spec_paths=self._cluster_specs)
        state_change_notifier = StateChangeNotifier(root_model.state_change_listener)

        root_controller = self._controller_hierarchy(state_change_notifier, root_model)
        root_controller.run(palette)

    def _controller_hierarchy(self, state_change_notifier, root_model):
        # RootController
        #   -> ClusterSelectionController
        #   -> NodePanelController
        #          -> ReceiveBufferController
        #          -> ClusterShellController
        #          -> SingleShellController

        cluster_selection_view = ClusterSelectionView(state_change_notifier)
        cluster_selection_controller = ClusterSelectionController(
            cluster_selection_view, root_model, state_change_notifier
        )

        receive_buffer_view = ReceiveBufferView(state_change_notifier)
        receive_buffer_controller = ReceiveBufferController(
            receive_buffer_view, root_model
        )

        cluster_shell_view = ClusterShellView(state_change_notifier)
        cluster_shell_controller = ClusterShellController(
            cluster_shell_view, root_model
        )

        single_shell_view = SingleShellView(state_change_notifier)
        single_shell_controller = SingleShellController(single_shell_view, root_model)

        node_panel_view = NodePanelView(
            state_change_notifier,
            receive_buffer_view,
            cluster_shell_view,
            single_shell_view,
        )
        node_panel_controller = NodePanelController(
            node_panel_view,
            root_model,
            cluster_shell_controller,
            single_shell_controller,
            receive_buffer_controller,
        )

        root_controller = RootController(
            state_change_notifier,
            root_model,
            cluster_selection_controller,
            node_panel_controller,
        )

        return root_controller



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


from typing import Callable, List
import urwid
from picsh.widgets.selectable_row import SelectableRow
from picsh.node import Node
from picsh.widgets.listbox_with_mouse_events import ListBoxWithMouseEvents
import os


class ClusterSelectionView:
    def __init__(self, state_change_notifier: Callable):
        self._state_change_notifier = state_change_notifier
        footer_text = "cluster specs dir /home/cborg/.picsh"
        self.column_headers = urwid.AttrMap(
            urwid.Columns(
                [
                    ("fixed", 80, urwid.Text("cluster yaml spec", align="left")),
                    ("fixed", 6, urwid.Text("", align="left")),
                ]
            ),
            "column_headers",
        )
        self.listbox_content = urwid.SimpleFocusListWalker([self.column_headers])
        listbox = urwid.ListBox(self.listbox_content)
        listbox = urwid.AttrWrap(listbox, "body")
        self._output_textbox = urwid.Text("This is some text. " * 200)
        textbox = self._output_textbox

        self._footer = urwid.Text(footer_text)
        header = urwid.AttrMap(urwid.Text("picsh >> select cluster"), "header_style")
        self._cols = urwid.Columns(
            [
                ("fixed", 46, listbox),
                (
                    "weight",
                    75,
                    ListBoxWithMouseEvents(urwid.SimpleFocusListWalker([textbox])),
                ),
            ]
        )
        self._frame = urwid.Frame(
            self._cols,
            header=header,
            footer=urwid.AttrMap(self._footer, "footer_style"),
        )
        self._focus = "nodes"

    def set_cluster_specs(self, cluster_spec_paths: List[str]):
        listbox_content = [self.column_headers]
        for idx, cluster_spec_path in enumerate(cluster_spec_paths):
            base_name = os.path.basename(cluster_spec_path)
            listbox_content.append(SelectableRow((base_name, ""), idx))
        self.listbox_content[:] = urwid.SimpleFocusListWalker(listbox_content)

    def show_cluster_spec(self, cluster_spec_paths: List[str]):
        spec_idx = self.get_selected_spec_idx()
        with open(cluster_spec_paths[spec_idx], "rt") as fh:
            buf = fh.read()
            self._output_textbox.set_text(buf)

    def outer_widget(self):
        return self._frame

    def log(self, msg):
        self._footer.set_text(msg)

    def toggle_focus(self):
        if self._focus == "specs":
            self._cols.set_focus(1)
            self._focus == "defn"
        else:
            self._cols.set_focus(0)
            self._focus == "specs"

    def get_selected_spec_idx(self):
        return self.listbox_content.focus - 1



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


from typing import Callable, List
import urwid
from picsh.widgets.selectable_row import SelectableRow
from picsh.node import Node
from picsh.widgets.listbox_with_mouse_events import ListBoxWithMouseEvents


class ReceiveBufferView:
    def __init__(self, state_change_notifier: Callable):
        self._state_change_notifier = state_change_notifier
        self._output_textbox = urwid.Text("  ")
        textbox = self._output_textbox
        self._outer_widget = ListBoxWithMouseEvents(
            urwid.SimpleFocusListWalker([textbox])
        )
        self._focus = "nodes"

    def on_node_list_modified(self, nodes: List[Node], node_idx):
        self._output_textbox.set_text(nodes[node_idx].recv_buf)

    def outer_widget(self):
        return self._outer_widget

    def header_nav_text(self):
        return "picsh >> response buffers"

    def footer_text(self):
        return "Ctrl s => cluster shell | Enter => single node ssh | Ctrl q => quit"



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import logging
import os
import asyncio
import urwid
import concurrent.futures
from collections import defaultdict
from picsh.views.cluster_shell_view import ClusterShellView
from picsh.views.view_names import ViewNames
from picsh.models.root_model import RootModel
from picsh.command_engine import CommandEngine
from picsh.controllers.base_controller import BaseController
import readline


class _TerminalSubProcess:
    def __init__(self):
        self._data_fd = None
        self._control_fd = None
        self.done = False

    def set_pipe(self, data_fd, control_fd):
        self._data_fd = data_fd
        self._control_fd = control_fd

    def run_input_loop(self):
        inp = ""
        while not self.done:
            try:
                inp = input("$ ")
                if inp.strip():
                    os.write(self._data_fd, inp.encode("utf-8"))
                    logging.info(f"cmd={inp}")
            except Exception as ex:
                os.write(self._data_fd, str(ex).encode("utf-8"))
            except KeyboardInterrupt as ex:
                print("Esc then Ctrl c to exit picsh")
        # os.write(self._control_fd, "done")


class ClusterShellController(BaseController):
    def __init__(self, view: ClusterShellView, root_model: RootModel):
        super().__init__()
        self._proc = None
        self.view: ClusterShellView = view
        self._model: RootModel = root_model
        for node in self._model.nodes:
            node.register_notify(self.on_command_output)
        self._command_queue = asyncio.Queue(-1)
        self._command_engine = CommandEngine(
            self._model.nodes, self.on_command_output, self._command_queue
        )
        self._terminal_proc = _TerminalSubProcess()
        self._activated = False

    def on_command_output(self):
        logging.info("got command output")
        self._repaint_notifier()

    def repaint(self):
        logging.info("repainting shell output")
        self.view.repaint_shell_output(self._model.nodes)

    def activate(self, **kwargs):
        if not self._activated:
            self.make_pipes(self._urwid_loop, self._aio_event_loop)
            self.start_command_loop(self._aio_event_loop)
            for node in self._model.nodes:
                node.register_notify(self.on_command_output)
            self.view.register_terminal_input_cmd(self._terminal_proc.run_input_loop)
            self.view.initialize_input_terminal(self._urwid_loop)
            self._activated = True
        self.view.set_initial_focus()

    def make_pipes(self, mainloop, aioloop):
        data_fd = mainloop.watch_pipe(self.on_data_pipe_data)
        control_fd = mainloop.watch_pipe(self.on_control_pipe_data)
        self._terminal_proc.set_pipe(data_fd, control_fd)

    def start_command_loop(self, aioloop):
        aioloop.create_task(self._command_engine.run_command_loop())

    def on_data_pipe_data(self, s_cmdline):
        s_cmdline = s_cmdline.decode("utf-8")
        self._reset_buffers()
        self._command_queue.put_nowait(s_cmdline)

    def _reset_buffers(self):
        for node in self._model.nodes:
            node.recv_buf = ""

    def on_control_pipe_data(self, data):
        raise urwid.ExitMainLoop()

    def quit(self):
        if self.view.term and self.view.term.pid:
            self.view.term.terminate()

    def handle_input_filter(self, keys, raw_input):
        new_view = None
        if "ctrl a" in keys:
            keys = []
            new_view = ViewNames.BUFFER_VIEW
        return keys, new_view

    def want_focus(self):
        return True



import pytest
from picsh.models.root_model import RootModel

def test_root_model():
    some_ref = ["aaa", "bbb"]
    root_model = RootModel(cluster_spec_paths=some_ref)
    root_model.state_change_listener({"cluster_spec_paths": ["zzz"]})
    assert some_ref == ["zzz"]
    assert root_model.cluster_spec_paths == ["zzz"]

    with pytest.raises(Exception):
        root_model.state_change_listener({"foo": "bar"})

if __name__ == "__main__":
    test_root_model()




import logging
import urwid
from picsh.controllers.cluster_selection_controller import ClusterSelectionController
from picsh.controllers.receive_buffer_controller import ReceiveBufferController
from picsh.controllers.single_shell_controller import SingleShellController
from picsh.models.root_model import RootModel
from picsh.views.node_panel_view import NodePanelView
from picsh.views.view_names import ViewNames
from picsh.controllers.base_controller import BaseController


class NodePanelController(BaseController):
    def __init__(
        self,
        node_panel_view: NodePanelView,
        root_model: RootModel,
        cluster_shell_controller: ClusterSelectionController,
        single_shell_controller: SingleShellController,
        receive_buffer_controller: ReceiveBufferController,
    ):
        super().__init__()
        self._urwid_loop = None
        self._aio_event_loop = None

        self.view: NodePanelView = node_panel_view
        self._root_model: RootModel = root_model

        self._cluster_shell_controller = cluster_shell_controller
        self._single_shell_controller = single_shell_controller
        self._receive_buffer_controller = receive_buffer_controller

        self.register_child_controller(cluster_shell_controller)
        self.register_child_controller(single_shell_controller)
        self.register_child_controller(receive_buffer_controller)

        self._active_controller = self._cluster_shell_controller
        self._active_view_name = ViewNames.CLUSTERSHELL_VIEW

        self._activated = False

    def handle_input(self, key):
        self._active_controller.handle_input(key)

    def switch_to(self, controller):
        self._active_controller = controller
        self._active_controller.activate(node_idx=self.view.get_selected_node_idx())
        self.view.set_focus(self._active_controller.want_focus())
        self.view.set_header_nav_text(self._active_controller.view.header_nav_text())
        self.view.set_footer_text(self._active_controller.view.footer_text())
        self.view.active_right_pane = self._active_controller.view
        self._urwid_loop.widget = self.view.outer_widget()

    def activate(self, **kwargs):
        if not self._activated:
            self.view.set_noderows(self._root_model.nodes)
            urwid.connect_signal(
                self.view.listbox_content, "modified", self.on_node_list_modified
            )
            self._activated = True
        self.switch_to(self._active_controller)

    def on_node_list_modified(self):
        node_idx = self.view.get_selected_node_idx()
        self.view.active_right_pane.on_node_list_modified(
            self._root_model.nodes, node_idx
        )

    def repaint(self):
        self.view.set_noderows(self._root_model.nodes)

    def handle_input_filter(self, keys, raw_input):
        if "ctrl q" in keys:
            keys = []
            new_view = ViewNames.EXIT_SCREEN
        else:
            keys, new_view = self._active_controller.handle_input_filter(keys, raw_input)
            if new_view == ViewNames.BUFFER_VIEW:
                self.switch_to(self._receive_buffer_controller)
            elif new_view == ViewNames.CLUSTERSHELL_VIEW:
                self.switch_to(self._cluster_shell_controller)
            elif new_view == ViewNames.SINGLESHELL_VIEW:
                self.switch_to(self._single_shell_controller)
            if new_view:
                self._active_view_name = new_view
        return keys, new_view




# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import urwid
import logging
from picsh.controllers.base_controller import BaseController
from picsh.models.root_model import RootModel
from picsh.views.receive_buffer_view import ReceiveBufferView
from picsh.views.view_names import ViewNames


class ReceiveBufferController(BaseController):
    def __init__(self, view: ReceiveBufferView, model: RootModel):
        super().__init__()
        self.view: ReceiveBufferView = view
        self._model: RootModel = model

    def activate(self, node_idx):
        self.view.on_node_list_modified(self._model.nodes, node_idx)

    def handle_input_filter(self, keys, raw_input):
        new_view = None
        if "ctrl s" in keys:
            keys = []
            new_view = ViewNames.CLUSTERSHELL_VIEW
        elif "enter" in keys:
            new_view = ViewNames.SINGLESHELL_VIEW
            keys = []
        return keys, new_view

    def want_focus(self):
        return False



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

from picsh.node import SSHTargetNode
from typing import List, Mapping, Dict, Any, MutableMapping

# store all application state in one place a la Flux
# https://youtu.be/nYkdrAPrdcw?t=635
# all state change occurs via a state_change_notifier


class RootModel:
    def __init__(self, nodes=None, cluster_spec_paths=None, node_selection_filter=""):
        self.nodes: List[SSHTargetNode] = nodes or []
        self.cluster_spec_paths: List[str] = cluster_spec_paths or []
        self.node_selection_filter: str = node_selection_filter

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name in self.__dict__:
            raise Exception("root_model: use state change listener to modify")
        else:
            super().__setattr__(__name, __value)

    def state_change_listener(self, update_dict: Mapping):
        for key, val in update_dict.items():
            if key not in self.__dict__:
                raise Exception("root_model: bad state variable")
            state_var = self.__dict__.get(key)
            if isinstance(state_var, dict):
                state_var.update(val)
            elif isinstance(state_var, list):
                state_var[:] = val
            else:
                self.__dict__[key] = val



#!/usr/bin/env python3
import logging
import warnings

warnings.filterwarnings("ignore", module="asyncssh\.crypto.*")
import sys
import os
import colorama
from picsh.controllers.root_controller import RootController
from picsh.util import get_logo, get_tagline
import glob
import yaml
from picsh.app import App


if sys.version_info.major < 3 or (sys.version_info.major == 3 and sys.version_info.minor < 6):
    print('Picsh only works with Python 3.6 and up. Your version is %s.x Exiting.' % sys.version_info.major)
    sys.exit(1)

if 'posix' not in os.name:
    print('Picsh has only been tested with on Linux/posix systems. Exiting')
    sys.exit(1)

def _get_cluster_specs():
    user_dir = os.path.expanduser("~")
    picsh_dir = os.path.join(user_dir, ".picsh")
    if not os.path.exists(picsh_dir):
        os.mkdir(picsh_dir)
    wildcardpath = os.path.join(picsh_dir, "*.yaml")
    cluster_specs = glob.glob(wildcardpath)
    ret = []
    for cluster_file in cluster_specs:
        if os.path.isfile(cluster_file):
            ret.append(cluster_file)
    return picsh_dir, ret


def _no_clusters_usage():
    print(
        "No cluster definitions found in [%s%s%s]"
        % (colorama.Fore.GREEN, picsh_dir, colorama.Style.RESET_ALL)
    )
    sample_defn = """
    cluster_name: kubedev
    login_user: root
    ssh_key_path: /home/devuser/keys/k8skey
    nodes:
    - ip: "10.0.0.1"
    - ip: "10.0.0.2"
    - ip: "10.0.0.3"
    """
    print(
        "%s%sSample cluster spec .yaml file:\n%s \nSee documentation for spec yaml.%s"
        % (
            colorama.Style.DIM,
            colorama.Fore.WHITE,
            sample_defn,
            colorama.Style.RESET_ALL,
        )
    )

def main():
    picsh_dir, cluster_specs = _get_cluster_specs()
    if len(cluster_specs) == 0:
        _no_clusters_usage()
        sys.exit(0)

    # TODO: if args.verbose
    logging.basicConfig(
        filename=os.path.join(picsh_dir, "picsh.log"),
        #encoding="utf-8",
        level=logging.INFO,
        format='%(asctime)s | %(levelname)8s | %(module)s : %(message)s',
        filemode="w"
    )
    logging.info("\n==> picsh started")

    App(cluster_specs).run()

    print(get_logo())
    print(colorama.Fore.CYAN)
    print(get_tagline())
    print(colorama.Fore.RESET)


if __name__ == "__main__":
    main()




# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

from typing import Callable, List

import urwid

from picsh.node import Node
from picsh.widgets.selectable_row import SelectableRow


class NodePanelView:
    def __init__(
        self,
        state_change_notifier: Callable,
        receive_buffer_view,
        cluster_shell_view,
        single_shell_view,
    ):
        self._state_change_notifier = state_change_notifier
        self._receive_buffer_view = receive_buffer_view
        self._cluster_shell_view = cluster_shell_view
        self._single_shell_view = single_shell_view

        self.active_right_pane = self._cluster_shell_view

        self.column_headers = urwid.AttrMap(
            urwid.Columns(
                [
                    ("fixed", 40, urwid.Text("    IP", align="left")),
                    ("fixed", 6, urwid.Text("bytes", align="left")),
                ]
            ),
            "column_headers",
        )
        self.listbox_content = urwid.SimpleFocusListWalker([self.column_headers])
        self._left_panel_listbox = urwid.AttrWrap(
            urwid.ListBox(self.listbox_content), "node_list"
        )

        self._header_text = urwid.Text(self.active_right_pane.header_nav_text())
        header = urwid.AttrMap(self._header_text, "header_style")

        self._output_textbox = urwid.Text("(no right panel rendered)")
        textbox = urwid.Filler(self._output_textbox, valign="top")
        right_panel = urwid.AttrMap(textbox, "node_text")

        self._footer = urwid.Text("")

        self._cols = urwid.Columns(
            [
                ("fixed", 47, urwid.LineBox(self._left_panel_listbox)),
                ("weight", 75, right_panel),
            ]
        )
        self._frame = urwid.Frame(
            self._cols,
            header=header,
            footer=urwid.AttrMap(self._footer, "footer_style"),
        )
        self._focus = "nodes"

    def set_noderows(self, nodes: List[Node]):
        listbox_content = [self.column_headers]
        for idx, node in enumerate(nodes):
            node_str = f"{str(node.idx).rjust(2)}] {node.get_ip()}"
            coldata = (node_str, str(len(node.recv_buf)))
            listbox_content.append(SelectableRow(coldata, node.idx))
        self.listbox_content[:] = urwid.SimpleFocusListWalker(listbox_content)

    def show_recv_buffer(self, nodes: List[Node]):
        node_idx = self.get_selected_node_idx()
        self._output_textbox.set_text(nodes[node_idx].recv_buf)

    def outer_widget(self):
        left_panel = (
            urwid.LineBox(
                self._left_panel_listbox, lline="", rline="", bline="", tline=""
            ),
            urwid.Columns.options("given", 47, True),
        )
        right_panel = (
            self.active_right_pane.outer_widget(),
            urwid.Columns.options("weight", 75, True),
        )
        self._frame.body.contents = [left_panel, right_panel]
        return self._frame

    def log(self, msg):
        self._footer.set_text(msg)

    def toggle_focus(self):
        if self._focus == "nodes":
            self._cols.set_focus(1)
            self._focus == "text"
        else:
            self._cols.set_focus(0)
            self._focus == "nodes"

    def get_selected_node_idx(self):
        return self.listbox_content.focus - 1

    def set_focus(self, on_right_panel: bool):
        self._cols.set_focus(1 if on_right_panel else 0)

    def set_header_nav_text(self, text: str):
        self._header_text.set_text(text)

    def set_footer_text(self, text: str):
        self._footer.set_text(text)



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

""" utility functions """

import logging
import colorama


def setup_logger(sname):
    logger = logging.getLogger(sname)

    console = logging.FileHandler("picsh.log")

    formatter = logging.Formatter("\rpicsh:%(asctime)s | %(message)s", "%H:%M:%S")
    console.setFormatter(formatter)

    logger.addHandler(console)
    logger.setLevel(logging.INFO)
    logger.propagate = False

    return logger


def get_logo():
    s_logo = """
                            ┌─┐@s╦╔═╗@e┌─┐┬ ┬
                            ├─┘@s║║  @e└─┐├─┤
                            ┴  @s╩╚═╝@e└─┘┴ ┴
    """
    s_logo = s_logo.replace("@s", colorama.Fore.CYAN)
    s_logo = s_logo.replace("@e", colorama.Fore.RESET)
    return s_logo


def get_tagline():
    s_github = """
                [.. Parallel Interactive Cluster Shell ..]
                    https://github.com/carlsborg/picsh 
    """
    return s_github



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


from typing import Callable, List

import urwid

from picsh.node import Node
from picsh.widgets.listbox_with_mouse_events import ListBoxWithMouseEvents


class ClusterShellView:
    def __init__(self, state_change_notifier: Callable):
        self._state_change_notifier = state_change_notifier
        self._footer = urwid.Text("Esc => node view | Alt-C reconnect all")
        self._listbox_content = urwid.SimpleListWalker([])
        self._listbox = ListBoxWithMouseEvents(self._listbox_content)
        self._terminal_input_cmd = None
        self._outer_widget = urwid.Filler(urwid.Text(""))
        self._node_selection_filter = ""
        self.term = None

    def set_selection_filter(self, selection_filter):
        self._node_selection_filter = selection_filter

    def register_terminal_input_cmd(self, func):
        self._terminal_input_cmd = func

    def terminal_command(self):
        self._terminal_input_cmd()

    def initialize_input_terminal(self, urwid_loop):
        self.term = urwid.Terminal(
            self.terminal_command, encoding="utf-8", main_loop=urwid_loop, escape_sequence="ctrl a"
        )
        term_attr = urwid.AttrMap(self.term, "bottom")
        # urwid.connect_signal(self.term, 'closed', quit)
        self._outer_widget = urwid.Pile([
                    self._listbox,
                    ("fixed", 5,  urwid.LineBox(
                        term_attr,
                        title="cluster shell",
                        title_attr="reveal_focus",
                        title_align="left",
                        bline="",
                        lline="",
                        rline="",
                     ))
            ]
        )

    def repaint_shell_output(self, nodes: List[Node]):
        list_contents = []
        for node in nodes:
            if not node.hide:
                separater = f"[{str(node.idx)}] {node.ip_addr}==>"
                list_contents.append(
                    urwid.AttrMap(urwid.Text(separater), "output_separater")
                )
                list_contents.append(urwid.Text(node.recv_buf))
        self._listbox_content[:] = urwid.SimpleListWalker(list_contents)

    def outer_widget(self):
        return self._outer_widget

    def log(self, msg):
        self._footer.set_text(msg)

    def on_node_list_modified(self, nodes: List[Node], node_idx):
        pass

    def set_initial_focus(self):
        self._outer_widget.set_focus(1)

    def header_nav_text(self):
        return "picsh >> cluster shell"

    def footer_text(self):
        return "Ctrl a => response buffers | Ctrl q => quit"
 



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


from typing import Optional
import asyncio, asyncssh
from functools import partial
from picsh.node import Node


class InteractiveClientSession(asyncssh.SSHClientSession):
    login_complete_guid = "B79D8677-F58A-4E09-B917-855A6619A951"

    def __init__(self, node: Node, notify_func):
        self._node = node
        self._login_guid_found = False
        self._notify_func = notify_func

    def data_received(self, data: str, datatype: asyncssh.DataType) -> None:
        self._node.recv_buf += data
        # if not self._login_guid_found:
        #     # surpress login banner.
        #     # Note: RFC-4254 reccomends the use of magic cookeis to surpress spurious
        #     # output when starting a subsystem via the shell "to distinguish it from
        #     # arbitrary output generated by shell initialization scripts, etc. This spurious
        #     # output from the shell may be filtered out either at the server or at the client"
        #     pos1 = self._node.recv_buf.find(InteractiveClientSession.login_complete_guid)
        #     if pos1 != -1:
        #         self._login_guid_found = True
        #         guid_len = len(InteractiveClientSession.login_complete_guid)
        #         self._node.recv_buf = self._node.recv_buf[pos1 + guid_len + 1:]
        if self._node.recv_buf:
            self._notify_func()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        err_str = "\npicsh: Connection lost. "
        if exc:
            err_str += "SSH session error: " + str(exc)
        self._node.recv_buf += err_str
        self._notify_func()


class CommandEngine:
    def __init__(self, nodes, notify_func, command_queue, loop=None):
        self._nodes = nodes
        self._notify_func = notify_func
        self._command_queue = command_queue
        self.done = False
        self._connections = {}
        self._sessions = {}
        self._using = []
        self._loop = loop

    def get_node_selection(self):
        return self._using

    async def _run_cmd_on_nodes(self, cmd, nodes):
        loop = self._loop or asyncio.get_event_loop()
        tasks = []
        for node in nodes:
            t = loop.create_task(node.run_cmd(cmd))
            tasks.append(t)

        for task in tasks:
            await task

    async def run_command_loop(self):
        while not self.done:
            cmd: str = await self._command_queue.get()
            try:
                if cmd.strip():
                    if cmd == "@*" or cmd == "@":
                        self._using = []
                        for node in self._nodes:
                            node.hide = False
                        self._notify_func()
                    elif cmd.startswith("@"):
                        target_nodes, cmd = self.target_nodes_from_at_str(cmd)
                        if not cmd.strip():
                            self._using = target_nodes
                            for node in self._nodes:
                                if node not in self._using:
                                    node.hide = True
                            self._notify_func()
                        else:
                            await self._run_cmd_on_nodes(cmd, target_nodes)
                    else:
                        target_nodes = self._using or self._nodes
                        await self._run_cmd_on_nodes(cmd, target_nodes)
            except Exception as ex:
                # TODO: send on control channel
                print("Oops. Error: " + str(ex))

    def target_nodes_from_at_str(self, cmd):
        parts = cmd.split()
        target_node_str = parts[0][1:]
        targets = target_node_str.split(",")
        target_idxs = [int(t.strip()) for t in targets]
        target_nodes = []
        for node in self._nodes:
            if node.idx in target_idxs:
                target_nodes.append(node)
        cmd = cmd[len(parts[0]) :]
        return target_nodes, cmd

    async def ensure_session(self, node):
        node_id = node.idx
        conn = self._connections.get(node_id)
        if not conn:
            conn = await self._make_connection(node)
            self._connections[node_id] = conn
        session_tuple = self._sessions.get(node_id)
        if not session_tuple:
            ssh_client_factory = partial(
                InteractiveClientSession, node, self._notify_func
            )
            session_tuple = await conn.create_session(ssh_client_factory)
            self._sessions[node_id] = session_tuple
        chan, sess = session_tuple
        return chan, sess

    async def _cmd(self, node, scmd: str):
        chan, sess = await self.ensure_session(node)
        chan.write(scmd + "\n")

    async def _make_connection(self, node: Node):
        conn, client = await asyncssh.create_connection(
            client_factory=None,
            host=node.ip_addr,
            username=node.login_user,
            known_hosts=None,
            options=asyncssh.SSHClientConnectionOptions(
                client_keys=node.ssh_key_path, username=node.login_user
            ),
        )
        return conn



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import logging
from typing import List, Callable
import os
import urwid
from picsh.node import Node


class _SingleShellTerminal:
    def __init__(self, node, loop):
        self._node = node
        self._loop = loop
        self.terminal_widget = urwid.Terminal(
            self._run_ssh_command,
            main_loop=loop,
            env=None,
            encoding="utf-8",
            escape_sequence="ctrl a",
        )
        self.exited = False

    def _run_ssh_command(self):
        key_opt = ""
        key_path = self._node.get_ssh_key_path()
        if key_path:
            key_opt = f"-i {self._node.get_ssh_key_path()}"
        ssh_cmd = f"TERM=linux ssh {key_opt} {self._node.get_login_user()}@{self._node.get_ip()}"
        logging.info(ssh_cmd)
        os.system(ssh_cmd)
        # self.terminal_widget.terminate()
        self.exited = True


class SingleShellView:
    def __init__(self, state_change_notifier: Callable):
        self._state_change_notifier = state_change_notifier
        self._terminals = {}  # node_idx : term
        self._outer_widget = urwid.Filler(urwid.Text(""))
        self._cur_node_ip = ""

    def outer_widget(self):
        return self._outer_widget

    def switch_to_terminal(self, node, urwid_loop):
        single_shell_term = self._terminals.get(node.idx)
        if single_shell_term and single_shell_term.terminal_widget.terminated:
            del self._terminals[node.idx]
            single_shell_term = None
        if not single_shell_term:
            single_shell_term = _SingleShellTerminal(node, urwid_loop)
            self._terminals[node.idx] = single_shell_term
        self._cur_node_ip = f"[{node.idx}] ({node.get_ip()})"
        self._outer_widget = self._terminals[node.idx].terminal_widget
        return

    def on_node_list_modified(self, nodes: List[Node], node_idx):
        pass

    def header_nav_text(self):
        return f"picsh >> ssh to node {self._cur_node_ip}"

    def footer_text(self):
        return "Ctrl a => response buffers | Ctrl s => cluster shell  | Ctrl q => quit"



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import warnings

warnings.filterwarnings("ignore", module="asyncssh\.crypto.*")
from typing import Optional
import asyncssh


class SSHTargetNode(asyncssh.SSHClientSession):
    def __init__(self):
        self.recv_buf = ""
        self.hide = False
        self._is_connected = False
        self._conn = None
        self._session_factory = None
        self._chan = None
        self._notify = None

    def session_factory(self):
        return self

    def register_notify(self, notify_func):
        self._notify = notify_func

    async def do_connect(self):
        self.recv_buf = f"connecting {self.get_login_user()}@{self.get_ip()} ..."
        self._notify()
        self._conn = await asyncssh.connect(
            self.get_ip(),
            known_hosts=None,
            options=asyncssh.SSHClientConnectionOptions(
                client_keys=self.get_ssh_key_path(),
                username=self.get_login_user(),
                password=self.get_password(),
            ),
        )
        self._chan, sess = await self._conn.create_session(self.session_factory)
        self.recv_buf += "\n...connected"
        self._notify()
        self.recv_buf = ""

    async def run_cmd(self, cmd):
        try:
            if not self._is_connected:
                await self.do_connect()
                self._is_connected = True
            self._chan.write(cmd + "\n")

        except Exception as ex:
            self.recv_buf = "picsh: exception encountered: " + str(ex)
            self._notify()

    def data_received(self, data: str, datatype: asyncssh.DataType) -> None:
        self.recv_buf += data
        # if not self._login_guid_found:
        #     # surpress login banner.
        #     # Note: RFC-4254 reccomends the use of magic cookeis to surpress spurious
        #     # output when starting a subsystem via the shell "to distinguish it from
        #     # arbitrary output generated by shell initialization scripts, etc. This spurious
        #     # output from the shell may be filtered out either at the server or at the client"
        #     pos1 = self._node.recv_buf.find(InteractiveClientSession.login_complete_guid)
        #     if pos1 != -1:
        #         self._login_guid_found = True
        #         guid_len = len(InteractiveClientSession.login_complete_guid)
        #         self._node.recv_buf = self._node.recv_buf[pos1 + guid_len + 1:]
        if self.recv_buf:
            self._notify()

    def connection_lost(self, exc: Optional[Exception]) -> None:
        err_str = "\npicsh: Connection lost. "
        if exc:
            err_str += "SSH session error: " + str(exc)
        self.recv_buf += err_str
        self._notify()

    def get_ip(self):
        raise NotImplementedError()

    def get_login_user(self):
        raise NotImplementedError()

    def get_ssh_key_path(self):
        raise NotImplementedError()

    def get_password(self):
        raise NotImplementedError()


class Node(SSHTargetNode):
    def __init__(self):
        super().__init__()
        self.ip_addr = ""
        self.login_user = ""
        self.password = ""
        self.ssh_key_path = ""
        self.name = ""
        self.idx = -1

    def get_ip(self):
        return self.ip_addr

    def get_login_user(self):
        return self.login_user

    def get_ssh_key_path(self):
        return self.ssh_key_path

    def get_password(self):
        return self.password

    def __str__(self):
        return f"{self.login_user}@{self.ip_addr}"

    def __repr__(self):
        return self.__str__()



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#
import logging
import urwid
from picsh.controllers.base_controller import BaseController
from picsh.state_change_notifier import StateChangeNotifier
from picsh.views.cluster_selection_view import ClusterSelectionView
from picsh.models.root_model import RootModel
from picsh.cluster_spec import ClusterSpec
from picsh.views.view_names import ViewNames


class ClusterSelectionController(BaseController):
    def __init__(
        self,
        view: ClusterSelectionView,
        model: RootModel,
        state_change_notifier: StateChangeNotifier,
    ):
        super().__init__()
        self.view: ClusterSelectionView = view
        self._model: RootModel = model
        self._state_change_notifier: StateChangeNotifier = state_change_notifier
        self.view.set_cluster_specs(model.cluster_spec_paths)
        urwid.connect_signal(self.view.listbox_content, "modified", self.on_modified)

    def on_modified(self):
        self.view.show_cluster_spec(self._model.cluster_spec_paths)

    def handle_input_filter(self, keys, raw_input):
        new_view = None
        if "enter" in keys:
            nodeidx = self.view.get_selected_spec_idx()
            cluster_spec_path = self._model.cluster_spec_paths[nodeidx]
            cluster_spec = ClusterSpec(cluster_spec_path)
            self._state_change_notifier.notify({"nodes": cluster_spec.nodes})
            keys = []
            new_view = ViewNames.NODE_PANEL_VIEW
        elif "ctrl c" in keys:
            new_view = ViewNames.EXIT_SCREEN
            keys = []
        return keys, new_view



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


import os
import yaml
from picsh.exceptions.picsh_exception import PicshException
from picsh.node import Node


class ClusterSpec:
    def __init__(self, path_to_yaml: str):
        if not os.path.exists(path_to_yaml):
            raise PicshException(
                f"Cluster spec [{path_to_yaml}] yaml file does not exist."
            )
        with open(path_to_yaml, "r") as fh:
            self.cluster_spec = yaml.safe_load(fh.read())
        self.nodes = self._nodes_from_cluster_spec(self.cluster_spec)
        self.name = "(missing)"

    def _nodes_from_cluster_spec(self, cluster):
        # globals
        hydrated_nodes = []
        login_user = cluster.get("login_user")
        key_path = cluster.get("ssh_key_path")
        self.name = cluster.get("cluster_name")
        for idx, node in enumerate(cluster.get("nodes")):
            n = Node()
            n.ip_addr = node["ip"]
            n.login_user = node.get("login_user") or login_user
            n.ssh_key_path = node.get("ssh_key_path") or key_path
            n.idx = idx
            hydrated_nodes.append(n)
        return hydrated_nodes



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import asyncio
import urwid
from picsh.controllers.base_controller import BaseController
from picsh.controllers.cluster_selection_controller import ClusterSelectionController
from picsh.controllers.node_panel_controller import NodePanelController
from picsh.models.root_model import RootModel
from picsh.views.view_names import ViewNames


class RootController(BaseController):
    def __init__(
        self,
        root_model: RootModel,
        pallette,
        cluster_selection_controller: ClusterSelectionController,
        node_panel_controller: NodePanelController,
    ):
        self._root_model = root_model

        self._cluster_selection_controller = cluster_selection_controller
        self._node_panel_controller = node_panel_controller

        self._active_controller = self._cluster_selection_controller
        self._active_view = self._cluster_selection_controller.view

    def switch_to(self, controller):
        self._active_controller = controller
        self._active_view = controller.view
        self._active_controller.activate()
        self._urwid_loop.widget = self._active_view.outer_widget()

    def handle_input(self, key):
        return self._active_controller.handle_input(key)

    def _input_filter(self, keys, raw_input):
        new_keys, new_view = self._active_controller.handle_input_filter(
            keys, raw_input
        )
        if new_view == ViewNames.NODE_PANEL_VIEW:
            self.switch_to(self._node_panel_controller)
        elif new_view == ViewNames.EXIT_SCREEN:
            self._cluster_selection_controller.quit()
            self._node_panel_controller.quit()
            raise urwid.ExitMainLoop()
        return new_keys

    def run(self, pallette):
        try:
            aio_event_loop = asyncio.get_event_loop()
            self._urwid_loop = urwid.MainLoop(
                self._active_view.outer_widget(),
                palette=pallette,
                unhandled_input=self.handle_input,
                input_filter=self._input_filter,
                event_loop=urwid.AsyncioEventLoop(loop=aio_event_loop),
            )
            self._node_panel_controller.set_urwid_loop(self._urwid_loop)
            self._node_panel_controller.set_aio_event_loop(aio_event_loop)
            self._urwid_loop.run()
        except KeyboardInterrupt:
            pass



from typing import List

class BaseController:
    def __init__(self):
        self._urwid_loop = None
        self._aio_event_loop = None
        self.child_controllers: List["BaseController"] = []
        self._repaint_notifier = None

    def register_child_controller(self, controller: "BaseController"):
        controller.set_repaint_notifier(self.repaint_notifer)
        self.child_controllers.append(controller)

    def set_urwid_loop(self, urwid_loop):
        self._urwid_loop = urwid_loop
        for c in self.child_controllers:
            c.set_urwid_loop(urwid_loop)

    def set_aio_event_loop(self, aio_event_loop):
        self._aio_event_loop = aio_event_loop
        for c in self.child_controllers:
            c.set_aio_event_loop(aio_event_loop)

    def activate(self, **kwargs):
        pass

    def handle_input(self, keys):
        return keys

    def handle_input_filter(self, keys, raw_input):
        return keys, None

    def switch_to(self):
        pass

    def repaint(self):
        pass

    def repaint_notifer(self):
        # child nodes can notify a parent node to repaint the tree
        self.repaint()
        for c in self.child_controllers:
            c.repaint()
        if self._urwid_loop:
            self._urwid_loop.draw_screen()

    def set_repaint_notifier(self, notifier):
        self._repaint_notifier = notifier

    def quit(self):
        for c in self.child_controllers:
            c.quit()



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import logging
from picsh.models.root_model import RootModel
from picsh.views.single_shell_view import SingleShellView
from picsh.views.view_names import ViewNames
from picsh.controllers.base_controller import BaseController


class SingleShellController(BaseController):
    def __init__(self, view: SingleShellView, model: RootModel):
        super().__init__()
        self.view: SingleShellView = view
        self._model: RootModel = model

    def handle_input(self, key):
        return key

    def handle_input_filter(self, keys, raw_input):
        new_view = None
        if "ctrl s" in keys:
            keys = []
            new_view = ViewNames.CLUSTERSHELL_VIEW
        elif "ctrl a" in keys:
            keys = []
            new_view = ViewNames.BUFFER_VIEW
        return keys, new_view

    def activate(self, node_idx):
        node = next(filter(lambda x: x.idx == node_idx, self._model.nodes), None)
        self.view.switch_to_terminal(node, self._urwid_loop)
        # urwid.connect_signal(
        #     self.view._terminals[node.idx].terminal_widget,
        #     "closed",
        #     functools.partial(self._switch_to, node),
        # )

    def want_focus(self):
        return True



#!/usr/bin/env python

import os
from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="picsh",
    long_description=read("Readme.md"),
)




# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#

import urwid


class SelectableRow(urwid.WidgetWrap):
    def __init__(self, contents, idx, on_select=None):
        self.contents = contents
        self.idx = idx
        self.on_select = on_select
        self._columns = urwid.Columns(
            [
                ("fixed", 40, urwid.Text(contents[0], align="left")),
                ("fixed", 6, urwid.Text(" " + contents[1], align="center")),
            ]
        )
        self._focusable_columns = urwid.AttrMap(self._columns, "", "reveal_focus")
        super(SelectableRow, self).__init__(self._focusable_columns)

    def selectable(self):
        return True

    def keypress(self, size, key):
        if self.on_select and key in ("enter",):
            self.on_select(self)
        return key

    def __repr__(self):
        return "%s(contents=%r)" % (self.__class__.__name__, self.contents)



# Copyright (c) Ran Dugal 2023
#
# This file is part of picsh
#
# Licensed under the GNU Affero General Public License v3, which is available at
# http://www.gnu.org/licenses/agpl-3.0.html
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Affero GPL for more details.
#


from typing import Mapping, Callable


class StateChangeNotifier:
    def __init__(self, state_change_listener: Callable):
        self._state_change_listener: Callable = state_change_listener

    def notify(self, state_update: Mapping):
        self._state_change_listener(state_update)
